# SimpleSpec
#
# Simulation integrated tests through examples.

require 'shrewd/simulation'

describe "simulation" do

  # Basic manufacturing example
  #
  # Uses a widget manufacturing example to test a simple simulation. Widgets
  # will arrive into the system at a uniformly randomized time. An available
  # machine will service each incoming widget. If there are not enough available
  # machines at the time a widget arrives, then the widget will sit in the
  # queue. The simulation will be run for a short amount of time, and its
  # results verified.
  context "manufacturing" do
    before :all do
      # Create system variables
      queue = Shrewd::Variable.new('Queue')
      machines = Shrewd::Variable.new('Machines', 'Machines', 5)
      variables = [queue, machines]

      # Create events
      widget_arrival = Shrewd::Event.new('Widget Arrival',
                                         'Widget arrives to manufacturer.')
      widget_arrival.action = Proc.new do |variables|
        variables[:queue].increment
      end

      start  = Shrewd::Event.new('Start Service', 'Start widget service')
      start.action = Proc.new do |variables|
        variables[:queue].decrement
        variables[:machines].decrement
      end

      finish = Shrewd::Event.new('Finish Service', 'Finish widget service')
      finish.action = Proc.new do |variables|
        variables[:machines].increment
      end

      # create processes
      arrival = Shrewd::Process.new('Arrival', 'Arrival of a Widget')
      arrival.delay = Proc.new { |variables| 10 }
      arrival.event = widget_arrival

      start_service = Shrewd::Process.new('Start Service',
                                          'Start Widget Service.')
      start_service.delay = Proc.new { |variables| 0 }
      start_service.conditions = Proc.new do |variables|
        variables[:machines] > 0 && variables[:queue] > 0
      end
      start_service.event = start

      service_widget = Shrewd::Process.new('Service', 'Servicing a Widget')
      service_widget.delay = Proc.new { |variables| 15 }
      service_widget.event = finish

      # Add processes to events
      widget_arrival.processes << arrival << start_service
      start.processes          << service_widget
      finish.processes         << start_service
      initial_processes        = [ arrival ]

      # Create simulation
      @simulation = Shrewd::Simulation.new(variables, initial_processes)
    end

    it "creates a simulation" do
      expect(@simulation).to be_an_instance_of Shrewd::Simulation
    end

    it "initializes processes" do
      expect(@simulation.initial_processes).to be_an_instance_of Array
      @simulation.initial_processes.each do |process|
        expect(process).to be_an_instance_of Shrewd::Process
      end
    end

    describe "when run" do
      before :all do
        @simulation_length = 1000
        @simulation.start(@simulation_length)
        @log = @simulation.print_log
      end

      it "does not last longer than simulation length" do
        expect(@log.last[:time]).to be <= @simulation_length
      end

      it "has more or equal widgets starts than finishes" do
        starts = 0
        finishes = 0

        @log.each do |entry|
          if entry[:event].name == 'Start Service'
            starts += 1
          elsif entry[:event].name == 'Finish Service'
            finishes += 1
          end

          expect(starts).to be >= finishes
        end

        expect(starts).to be > 0
        expect(finishes).to be > 0
        expect(starts).to be >= finishes
      end

      it "has more or equal widget arrivals than starts" do
        starts = 0
        arrivals = 0

        @log.each do |entry|
          if entry[:event].name == 'Start Service'
            starts += 1
          elsif entry[:event].name == 'Widget Arrival'
            arrivals += 1
          end

          expect(arrivals).to be >= starts
        end

        expect(arrivals).to be > 0
        expect(starts).to be > 0
        expect(arrivals).to be >= starts
      end

      it "executes events in order" do
        clock = 0

        @log.each do |entry|
          expect(entry[:time]).to be >= clock
          clock = entry[:time]
        end
      end

      after :all do
        @log.each do |entry|
          puts "time #{entry[:time]} - #{entry[:event].name}"
        end
      end
    end
  end
end
