# SimulationSpec
#
# Tests for the Shrewd::Simulation class

require 'shrewd/simulation'

describe Shrewd::Simulation do

  # Shared variables
  let(:state_variables) { [ Shrewd::Variable.new('Queue'),
                            Shrewd::Variable.new('Employees', '', 5)] }
  let(:process_length) { 20 }
  let(:simulation_length) { 1000 }
  let(:initial_processes) do
    process = Shrewd::Process.new('Process1')
    process.delay = Proc.new { process_length }

    event = Shrewd::Event.new('Arrival')
    event.processes << process
    process.event = event
    
    [ process ]
  end

  # Create a simulation instance before each test
  before :each do
    @simulation = Shrewd::Simulation.new(state_variables, initial_processes)
  end

  # Validate new simulation instance
  describe "#new" do
    it "should be a simulation" do
      expect(@simulation).to be_an_instance_of Shrewd::Simulation
    end

    it "instantiates variables correctly" do
      variables = @simulation.instance_variable_get(:@variables)
      expect(variables).to be_an_instance_of Hash

      variables.values.each do |variable|
        expect(variable).to be_an_instance_of Shrewd::Variable
      end

      variables.each do |key, value|
        expect(key).to eql value.name.downcase.gsub(/\s+/,"_").to_sym
      end

      expect(@simulation.initial_processes).to be_an_instance_of Array
      expect(@simulation.initial_processes).to eql initial_processes

      @simulation.initial_processes.each do |process|
        expect(process).to be_an_instance_of Shrewd::Process
      end
    end
  end

  describe "#start" do
    before :each do
      @simulation.start(simulation_length)
    end

    it "generates a start event" do
      expect(@simulation.print_log.size).to eql (simulation_length/process_length) + 1
      expect(@simulation.print_log.first[:event].name).to eql 'Start'
    end
  end

  describe "#print_log" do
    it "generates a log" do
      expect(@simulation.print_log).to be_an_instance_of Array
    end
  end

  describe "#create_initial_event" do
    before :each do
      @simulation.send(:create_initial_event)
      @queue = @simulation.instance_variable_get(:@queue)
    end

    it "adds initial event to the queue" do
      expect(@queue.size).to eql 1
    end

    it "creates initial event" do
      initial_event = @queue.first[:event]
      expect(initial_event).to be_an_instance_of Shrewd::Event
      expect(initial_event.name).to eql 'Start'
    end

    it "adds initial processes to initial event" do
      initial_event = @queue.first[:event]
      expect(initial_event.processes.size).to be > 0
      expect(initial_event.processes).to eql initial_processes
    end
  end

  describe "#add_to_queue" do
    let(:first_name) { 'First' }
    let(:first_time) { 5 }
    let(:second_name) { 'Second' }
    let(:second_time) { 10 }
    let(:third_name) { 'Third' }
    let(:third_time) { 7 }
    let(:fourth_name) { 'Fourth' }
    let(:fourth_time) { 3 }

    before :each do
      @initial_event = Shrewd::Event.new(first_name)
      @simulation.send(:add_to_queue, { time: first_time, event: @initial_event })
      @queue = @simulation.instance_variable_get(:@queue)
    end

    it "adds an event to the queue" do
      expect(@queue).to be_an_instance_of Array
      expect(@queue).to_not be_empty
      expect(@queue.first[:event].name).to eql first_name
      expect(@queue.first[:time]).to eql first_time
    end

    # Test base binary search
    context "multiples events" do
      before :each do
        @second_event = Shrewd::Event.new(second_name)
        @simulation.send(:add_to_queue, { time: second_time,
                                          event: @second_event })
      end

      it "adds second event to the queue" do
        expect(@queue.size).to eql 2
        expect(@queue[1][:event].name).to eql second_name
        expect(@queue[1][:time]).to eql second_time
      end

      context "with third and fourth event" do
        before :each do
          @third_event = Shrewd::Event.new(third_name)
          @fourth_event = Shrewd::Event.new(fourth_name)

          @simulation.send(:add_to_queue, { time: third_time,
                                            event: @third_event })
          @simulation.send(:add_to_queue, { time: fourth_time,
                                            event: @fourth_event })
        end

        it "adds third and fourth events to the queue" do
          expect(@queue.size).to eql 4
        end

        it "places events in order" do
          time = 0
          @queue.each do |item|
            expect(item[:time]).to be > time
            time = item[:time]
          end
        end
      end
    end
  end

  describe "#reset" do
    before :each do
      @simulation.start(1000)
      @simulation.reset
    end

    it "resets system state" do
      log        = @simulation.print_log
      queue      = @simulation.instance_variable_get(:@queue)
      clock      = @simulation.instance_variable_get(:@clock)
      variables  = @simulation.instance_variable_get(:@variables)
      initial    = @simulation.instance_variable_get(:@initial_state)

      expect(log).to be_empty
      expect(queue).to be_empty
      expect(clock).to eql 0
      expect(variables).to eql initial
    end

  end
end
