# VariableSpec
#
# Tests for the Shrewd::Variable class

require 'shrewd/entities/variable'

describe Shrewd::Variable do

  # Shared variables
  let(:name) {  'Queue' }
  let(:description) { 'Number of people in line.' }
  let(:initial_value) { 5 }

  describe "initialization" do

    # Common examples for validating a variable
    shared_examples "valid variable" do

      it "should be a variable" do
        expect(@variable).to be_an_instance_of Shrewd::Variable
      end

      it "should have the correct properties" do
        expect(@variable.name).to eql name
        expect(@variable.description).to eql description
      end

    end

    context "when default value provided" do

      before :each do
        @variable = Shrewd::Variable.new(name, description, initial_value)
      end

      include_examples "valid variable"

      it "should have the provided default value" do
        expect(@variable.value).to eql initial_value
      end

    end

    context "when no default value provided" do

      before :each do
        @variable = Shrewd::Variable.new(name, description)
      end

      include_examples "valid variable"

      it "should have a default of 0" do
        expect(@variable.value).to eql 0
      end

    end

  end

  describe "increment" do

    let(:positive_amount) { 4 }
    let(:negative_amount) { -2 }

    before :each do
      @variable = Shrewd::Variable.new(name, description, initial_value)
    end

    context "by default amount" do

      before :each do
        @variable.increment
      end

      it "should add one to initial value" do
        expected_value = initial_value + 1
        expect(@variable.value).to eql expected_value
      end
    end

    context "by positive amount" do

      before :each do
        @variable.increment positive_amount
      end

      it "should add positive amount to initial value" do
        expected_value = initial_value + positive_amount
        expect(@variable.value).to eql expected_value
      end

    end

    context "by negative amount" do

      before :each do
        @variable.increment negative_amount
      end

      it "should subtract negative amount from initial value" do
        expected_value = initial_value + negative_amount
        expect(@variable.value).to eql expected_value
      end

    end
  end

  describe "decrement" do

    let(:positive_amount) { 2 }
    let(:negative_amount) { -2 }

    before :each do
      @variable = Shrewd::Variable.new(name, description, initial_value)
    end

    context "by default amount" do

      before :each do
        @variable.decrement
      end

      it "should subtract one from initial value" do
        expected_value = initial_value - 1
        expect(@variable.value).to eql expected_value
      end
    end

    context "by provided positive amount" do

      before :each do
        @variable.decrement positive_amount
      end

      it "should subtract positive amount from initial value" do
        expected_value = initial_value - positive_amount
        expect(@variable.value).to eql expected_value
      end

    end

    context "by provided negative amount" do

      before :each do
        @variable.decrement negative_amount
      end

      it "should add negative amount to initial value" do
        expected_value = initial_value - negative_amount
        expect(@variable.value).to eql expected_value
      end

    end
  end

  describe "compare" do

    context "with self" do
      before :each do
        @variable1 = Shrewd::Variable.new(name, description, initial_value)
        @variable2 = @variable1
      end

      it "should be equal" do
        expect(@variable1 == @variable2.value).to be true
      end
    end

    context "with variable with same value" do
      before :each do
        @variable1 = Shrewd::Variable.new(name, description, initial_value)
        @variable2 = Shrewd::Variable.new(name, description, initial_value)
      end

      it "should be equal" do
        expect(@variable1 == @variable2.value).to be true
      end
    end

    context "with a variable with different value" do
      let(:different_value) { initial_value + 3 }

      before :each do
        @variable1 = Shrewd::Variable.new(name, description, initial_value)
        @variable2 = Shrewd::Variable.new(name, description, different_value)
      end

      it "should not be equal" do
        expect(@variable1 == @variable2.value).to be false
      end

    end

  end
end
